# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

# This file is here so that pytest won't complain that it cannot find tests.
# Remove once we have actual tests.

import antsibull


def test_dummy():
    pass
