{#
# Copyright (C) Ansible Project
# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+
# (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
#}
Hello everyone,

We're happy to announce the release of the Ansible {{ version }} package{% if is_prerelease %} pre-release{% endif %}!

Ansible {{ version }} depends on ansible-core {{ core_version }} and includes a curated set of Ansible collections that provide a vast number of modules, plugins, and roles.{% if is_major_release and is_prerelease %} This is a pre-release of Ansible {{ major_version }}.{% elif is_major_release %} This is the first stable release of Ansible {{ major_version }}.{% elif is_prerelease %} This is a pre-release of Ansible.{% endif %}

{{ "How to get it" | email_heading }}

This {% if is_prerelease %}pre-{% endif %}release is available on PyPI and can be installed with pip:

```console
python3 -m pip install ansible=={{ version }} --user
```

The sources for this release can be found here:

Release tarball:

{{ release_tarball.url }}

SHA256:

{{ release_tarball.sha256sum }}

Wheel package:

{{ release_wheel.url }}

SHA256:

{{ release_wheel.sha256sum }}


{{ "Some important details" | email_heading }}

ansible-core is a separate package on which ansible depends. `pip install ansible` installs `ansible-core`, but it can also be installed independently of the ansible package.

Collections that have opted to join the Ansible {{ major_version }} unified changelog will have an entry on this page:

{{ build_data_path }}/CHANGELOG-v{{ major_version }}.md

For collections which have not opted-in to the unified changelog, you may find more information on

https://docs.ansible.com/ansible/latest/collections

or on the collection source repository. For example, the community.crypto collection is available at

https://docs.ansible.com/ansible/latest/collections/community/crypto/index.html

and you can find a link to the source repository under the “Repository (Sources)” button.

The changelog for ansible-core {{ core_major_version }} installed by this release of Ansible {{ major_version }} can be found here:

https://github.com/ansible/ansible/blob/v{{ core_major_version }}/changelogs/CHANGELOG-v{{ core_major_version }}.rst

{{ ("What's the schedule for new Ansible releases after " ~ version ~ "?") | email_heading }}

The next release roadmap can be found at

https://docs.ansible.com/ansible/devel/roadmap/ansible_roadmap_index.html

The Ansible community package release schedule follows the Ansible Core release schedule, including, for example, delays for holidays. This means Ansible releases happen every four weeks through most of the year, but release dates may be delayed when Ansible Core releases are.

Subscribe to the Bullhorn for all future release dates, announcements, and Ansible contributor community news.

Visit this link to subscribe: https://bit.ly/subscribe-bullhorn

You can find all past Bullhorn issues on the Ansible Community Forum at:

https://forum.ansible.com/c/news/bullhorn/17

Join the new Ansible Community Forum to follow along and participate in all the project and release related discussions and announcements. Feel free to share your thoughts, feedback, ideas and concerns there.

Register here to join the Ansible Forum:

https://forum.ansible.com

{{ "Porting Help" | email_heading }}

A unified porting guide for collections that have opted in is available here:

https://docs.ansible.com/ansible/devel/porting_guides/porting_guide_{{ major_version }}.html

{{ ("Getting collection updates from Ansible " ~ major_version ~ " with older releases of ansible-core") | email_heading }}

Ansible {{ major_version }} depends on ansible-core {{ core_major_version }}. Depending on your needs, you can get collection updates as they ship in the Ansible “batteries included” package while continuing to use older versions of ansible-core.

See the ansible-galaxy requirements file based on the collections from Ansible {{ major_version }} for this use case:

{{ build_data_path }}/galaxy-requirements.yaml

After you download the requirements file, you can install the collections by running the following command:

```console
ansible-galaxy collection install -r galaxy-requirements.yaml
```

On behalf of the Ansible community, thank you and happy automating!

Cheers,
Ansible Release Management Working Group
