# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

"""
Minimal setup.py for compatibility with older tools.
Configuration is in setup.cfg.
Users are recommended to use tools like pip and build and the PEP 517 interface
instead of setuptools' deprecated setup.py interface.
"""

from setuptools import setup

setup()
