{#
# Copyright (C) Ansible Project
# Copyright (C) 2024 Maxwell G <maxwell@gtmx.me>
# SPDX-License-Identifier: GPL-3.0-or-later
# GNU General Public License v3.0+
# (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
#}
 Ansible {{ version }} package{% if is_prerelease %} (pre-release){% endif %} is here! ❤️
🔗<FORUM LINK>
💽You can install it by running the following command:

```
python3 -m pip install ansible=={{ version }} --user
```

➡️ Check [Release Notes📦️🗒️]({{ build_data_path }}/CHANGELOG-v{{ major_version }}.md) and [Ansible {{ major_version }} Porting Guide](https://docs.ansible.com/ansible/devel/porting_guides/porting_guide_{{ major_version }}.html) for more details!
