# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Ansible Project, 2020

ansible ({{ version }}-1) unstable; urgency=medium

  * Ansible {{ version }} release

 -- Red Hat, Inc. <info@ansible.com>  {{ date }}
